import baseLogger from './logger';
import EventEmitter from './EventEmitter';
import postProcessor from './postProcessor';
import * as compat from './compatibility/v1';
import * as utils from './utils';

var Translator = function (_EventEmitter) {
  babelHelpers.inherits(Translator, _EventEmitter);

  function Translator(services) {
    var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
    babelHelpers.classCallCheck(this, Translator);

    var _this = babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Translator).call(this));

    utils.copy(['resourceStore', 'languageUtils', 'pluralResolver', 'interpolator', 'backendConnector'], services, _this);

    _this.options = options;
    _this.logger = baseLogger.create('translator');
    return _this;
  }

  babelHelpers.createClass(Translator, [{
    key: 'changeLanguage',
    value: function changeLanguage(lng) {
      if (lng) this.language = lng;
    }
  }, {
    key: 'exists',
    value: function exists(key) {
      var options = arguments.length <= 1 || arguments[1] === undefined ? { interpolation: {} } : arguments[1];

      if (this.options.compatibilityAPI === 'v1') {
        options = compat.convertTOptions(options);
      }

      return this.resolve(key, options) !== undefined;
    }
  }, {
    key: 'extractFromKey',
    value: function extractFromKey(key, options) {
      var nsSeparator = options.nsSeparator || this.options.nsSeparator;
      if (nsSeparator === undefined) nsSeparator = ':';

      var namespaces = options.ns || this.options.defaultNS;
      if (nsSeparator && key.indexOf(nsSeparator) > -1) {
        var parts = key.split(nsSeparator);
        namespaces = parts[0];
        key = parts[1];
      }
      if (typeof namespaces === 'string') namespaces = [namespaces];

      return {
        key: key,
        namespaces: namespaces
      };
    }
  }, {
    key: 'translate',
    value: function translate(keys) {
      var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

      if ((typeof options === 'undefined' ? 'undefined' : babelHelpers.typeof(options)) !== 'object') {
        options = this.options.overloadTranslationOptionHandler(arguments);
      } else if (this.options.compatibilityAPI === 'v1') {
        options = compat.convertTOptions(options);
      }

      // non valid keys handling
      if (keys === undefined || keys === null || keys === '') return '';
      if (typeof keys === 'number') keys = String(keys);
      if (typeof keys === 'string') keys = [keys];

      // return key on CIMode
      var lng = options.lng || this.language;
      if (lng && lng.toLowerCase() === 'cimode') return keys[keys.length - 1];

      // separators
      var keySeparator = options.keySeparator || this.options.keySeparator || '.';

      // get namespace(s)

      var _extractFromKey = this.extractFromKey(keys[keys.length - 1], options);

      var key = _extractFromKey.key;
      var namespaces = _extractFromKey.namespaces;

      var namespace = namespaces[namespaces.length - 1];

      // resolve from store
      var res = this.resolve(keys, options);

      var resType = Object.prototype.toString.apply(res);
      var noObject = ['[object Number]', '[object Function]', '[object RegExp]'];
      var joinArrays = options.joinArrays !== undefined ? options.joinArrays : this.options.joinArrays;

      // object
      if (res && typeof res !== 'string' && noObject.indexOf(resType) < 0 && !(joinArrays && resType === '[object Array]')) {
        if (!options.returnObjects && !this.options.returnObjects) {
          this.logger.warn('accessing an object - but returnObjects options is not enabled!');
          return this.options.returnedObjectHandler ? this.options.returnedObjectHandler(key, res, options) : 'key \'' + key + ' (' + this.language + ')\' returned an object instead of string.';
        }

        var copy = resType === '[object Array]' ? [] : {}; // apply child translation on a copy

        for (var m in res) {
          copy[m] = this.translate('' + key + keySeparator + m, babelHelpers.extends({ joinArrays: false, ns: namespaces }, options));
        }
        res = copy;
      }
      // array special treatment
      else if (joinArrays && resType === '[object Array]') {
          res = res.join(joinArrays);
          if (res) res = this.extendTranslation(res, key, options);
        }
        // string, empty or null
        else {
            var usedDefault = false,
                usedKey = false;

            // fallback value
            if (!this.isValidLookup(res) && options.defaultValue) {
              usedDefault = true;
              res = options.defaultValue;
            }
            if (!this.isValidLookup(res)) {
              usedKey = true;
              res = key;
            }

            // save missing
            if (usedKey || usedDefault) {
              this.logger.log('missingKey', lng, namespace, key, res);

              if (this.options.saveMissing) {
                var lngs = [];
                if (this.options.saveMissingTo === 'fallback' && this.options.fallbackLng && this.options.fallbackLng[0]) {
                  for (var i = 0; i < this.options.fallbackLng.length; i++) {
                    lngs.push(this.options.fallbackLng[i]);
                  }
                } else if (this.options.saveMissingTo === 'all') {
                  lngs = this.languageUtils.toResolveHierarchy(options.lng || this.language);
                } else {
                  //(this.options.saveMissingTo === 'current' || (this.options.saveMissingTo === 'fallback' && this.options.fallbackLng[0] === false) ) {
                  lngs.push(options.lng || this.language);
                }

                if (this.options.missingKeyHandler) {
                  this.options.missingKeyHandler(lngs, namespace, key, res);
                } else if (this.backendConnector && this.backendConnector.saveMissing) {
                  this.backendConnector.saveMissing(lngs, namespace, key, res);
                }

                this.emit('missingKey', lngs, namespace, key, res);
              }
            }

            // extend
            res = this.extendTranslation(res, key, options);

            // append namespace if still key
            if (usedKey && res === key && this.options.appendNamespaceToMissingKey) res = namespace + ':' + key;

            // parseMissingKeyHandler
            if (usedKey && this.options.parseMissingKeyHandler) res = this.options.parseMissingKeyHandler(res);
          }

      // return
      return res;
    }
  }, {
    key: 'extendTranslation',
    value: function extendTranslation(res, key, options) {
      var _this2 = this;

      if (options.interpolation) this.interpolator.init(options);

      // interpolate
      var data = options.replace && typeof options.replace !== 'string' ? options.replace : options;
      if (this.options.interpolation.defaultVariables) data = babelHelpers.extends({}, this.options.interpolation.defaultVariables, data);
      res = this.interpolator.interpolate(res, data);

      // nesting
      res = this.interpolator.nest(res, function () {
        for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
          args[_key] = arguments[_key];
        }

        return _this2.translate.apply(_this2, args);
      }, options);

      if (options.interpolation) this.interpolator.reset();

      // post process
      var postProcess = options.postProcess || this.options.postProcess;
      var postProcessorNames = typeof postProcess === 'string' ? [postProcess] : postProcess;

      if (res !== undefined && postProcessorNames && postProcessorNames.length && options.applyPostProcessor !== false) {
        res = postProcessor.handle(postProcessorNames, res, key, options, this);
      }

      return res;
    }
  }, {
    key: 'resolve',
    value: function resolve(keys) {
      var _this3 = this;

      var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

      var found = void 0;

      if (typeof keys === 'string') keys = [keys];

      // forEach possible key
      keys.forEach(function (k) {
        if (_this3.isValidLookup(found)) return;

        var _extractFromKey2 = _this3.extractFromKey(k, options);

        var key = _extractFromKey2.key;
        var namespaces = _extractFromKey2.namespaces;

        if (_this3.options.fallbackNS) namespaces = namespaces.concat(_this3.options.fallbackNS);

        var needsPluralHandling = options.count !== undefined && typeof options.count !== 'string';
        var needsContextHandling = options.context !== undefined && typeof options.context === 'string' && options.context !== '';

        var codes = options.lngs ? options.lngs : _this3.languageUtils.toResolveHierarchy(options.lng || _this3.language);

        namespaces.forEach(function (ns) {
          if (_this3.isValidLookup(found)) return;

          codes.forEach(function (code) {
            if (_this3.isValidLookup(found)) return;

            var finalKey = key;
            var finalKeys = [finalKey];

            var pluralSuffix = void 0;
            if (needsPluralHandling) pluralSuffix = _this3.pluralResolver.getSuffix(code, options.count);

            // fallback for plural if context not found
            if (needsPluralHandling && needsContextHandling) finalKeys.push(finalKey + pluralSuffix);

            // get key for context if needed
            if (needsContextHandling) finalKeys.push(finalKey += '' + _this3.options.contextSeparator + options.context);

            // get key for plural if needed
            if (needsPluralHandling) finalKeys.push(finalKey += pluralSuffix);

            // iterate over finalKeys starting with most specific pluralkey (-> contextkey only) -> singularkey only
            var possibleKey = void 0;
            while (possibleKey = finalKeys.pop()) {
              if (_this3.isValidLookup(found)) continue;
              found = _this3.getResource(code, ns, possibleKey, options);
            }
          });
        });
      });

      return found;
    }
  }, {
    key: 'isValidLookup',
    value: function isValidLookup(res) {
      return res !== undefined && !(!this.options.returnNull && res === null) && !(!this.options.returnEmptyString && res === '');
    }
  }, {
    key: 'getResource',
    value: function getResource(code, ns, key) {
      var options = arguments.length <= 3 || arguments[3] === undefined ? {} : arguments[3];

      return this.resourceStore.getResource(code, ns, key, options);
    }
  }]);
  return Translator;
}(EventEmitter);

export default Translator;